#!/usr/bin/perl -w
#
# Copyright 2012, Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This example illustrates how to handle 'captcha required' authorization
# errors. Refer to the best practices guide on how to avoid this error:
#
# http://code.google.com/apis/adwords/docs/guides/bestpractices.html#auth_tokens
#
# Author: David Torres <api.davidtorres@gmail.com>

use strict;
use lib "../../../lib";

use Google::Ads::AdWords::Client;
use Google::Ads::Common::AuthError;
use Google::Ads::Common::AuthToken;
use Google::Ads::Common::CaptchaRequiredError;

use Cwd qw(abs_path);
use POSIX;

use constant RETRIES => 500;

# Example main subroutine.
sub handle_captcha_challenge {
  my $client = shift;

  my $retry = RETRIES;
  # Brute forcing the generation of a captcha challenge.
  while ($retry--) {
    eval {
      Google::Ads::Common::AuthToken::get_token({
        email => $client->get_email(),
        password => $client->get_password(),
        service => 'adwords'
      });
    };
    if ($@) {
      last;
    }
  }

  if (!$@) {
    print "Failed to trigger a captcha challenge.\n";
    return;
  }

  if ($@->isa("Google::Ads::Common::CaptchaRequiredError")) {
    my $captcha_error = $@;

    # Captured a captcha required error.
    print "A captcha challenge error has ocurred. To recover open the " .
          "following URL in a browser:\n" .
          $captcha_error->get_image();
    print "\nEnter the captcha code and press ENTER to retry: ";
    my $code = <STDIN>;

    eval {
      my $token = Google::Ads::Common::AuthToken::get_token({
        email => $client->get_email(),
        password => $client->get_password(),
        service => 'adwords',
        captcha_token => $captcha_error->get_token(),
        captcha_code => $code
      });
      print "Successfully retrived token: " . $token. "\n";
    };
    if ($@) {
      print "Invalid captcha code given.\n";
    }
  } elsif ($@->isa("Google::Ads::Common::AuthError")) {
    # Captured a different authentication error.
    print "A different authentication error has occurred, please check your " .
          "client credentials.\n";
    print $@;
  }

  return 1;
}

# Don't run the example if the file is being included.
if (abs_path($0) ne abs_path(__FILE__)) {
  return 1;
}

# Get AdWords Client, credentials will be read from ~/adwords.properties.
my $client = Google::Ads::AdWords::Client->new({version => "v201109"});

# Call the example
handle_captcha_challenge($client);
