# Copyright 2011, Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

package Google::Ads::Common::AuthToken;

use strict;
use version;

# The following needs to be on one line because CPAN uses a particularly hacky
# eval() to determine module versions.
use Google::Ads::Common::Constants; our $VERSION = ${Google::Ads::Common::Constants::VERSION};

use Google::Ads::Common::AuthError;
use Google::Ads::Common::CaptchaRequiredError;

use LWP::UserAgent;
use URI::Escape;

# These constants should be fairly, well, constant. But if anything does change
# with the ClientLogin mechanism, this should be updated.
use constant ACCOUNT_TYPE => "GOOGLE";
use constant DEFAULT_SERVER => "https://www.google.com";
use constant AUTH_URL => "%s/accounts/ClientLogin";
use constant CAPTCHA_URL_PATH => "/accounts/";

our $user_agent = LWP::UserAgent->new();

sub set_user_agent {
  $user_agent = shift;
}

# Uses the ClientLogin web service to request an auth token.
sub get_token(%) {
  my ($params) = @_;
  my $service = uri_escape($params->{service})
    or die("Required hash parameter 'service' not provided.");
  my $email = uri_escape($params->{email})
    or die("Required hash parameter 'email' not provided.");
  my $password = uri_escape($params->{password})
    or die("Required hash parameter 'password' not provided.");
  my $server = $params->{server} ? $params->{server} : DEFAULT_SERVER;
  my $captcha_token = $params->{captcha_token} ?
      uri_escape($params->{captcha_token}) : undef;
  my $captcha_code = $params->{captcha_code} ?
      uri_escape($params->{captcha_code}) : undef;

  # The ClientLogin interface is documented at
  # http://code.google.com/apis/accounts/docs/AuthForInstalledApps.html
  my $data = sprintf("accountType=%s&Email=%s&Passwd=%s&service=%s",
                     ACCOUNT_TYPE, $email, $password, $service);
  if ($captcha_token && $captcha_code) {
    $data = sprintf("%s&logintoken=%s&logincaptcha=%s",
                    $data, $captcha_token, $captcha_code);
  }

  # my $userAgent = LWP::UserAgent->new();
  $user_agent->env_proxy();
  $user_agent->agent(sprintf("%s: %s", __PACKAGE__, $0));

  my $request = HTTP::Request->new(POST => sprintf(AUTH_URL, $server));
  $request->content_type("application/x-www-form-urlencoded");
  $request->content($data);

  my $response = $user_agent->request($request);
  my $content = $response->content();

  if ($response->is_success()) {

    # Sample response body:
    #   SID=DQAAAGgA...7Zg8CTN
    #   LSID=DQAAAGsA...lk8BBbG
    #   Auth=DQAAAGgA...dk3fA5N
    # We only care about the Auth token.
    if ($content =~ /Auth=(.+)$/) {
      return $1;
    }
  }

  my %content_hash = __content_to_hash($content);

  if ($content_hash{Error} eq "CaptchaRequired") {
    my $captcha_image = $server . CAPTCHA_URL_PATH . $content_hash{CaptchaUrl};

    die(Google::Ads::Common::CaptchaRequiredError->new({
      token => $content_hash{CaptchaToken},
      image => $captcha_image,
      url => $content_hash{Url},
      message => $response->message(),
      code => $response->code(),
      content => $response->content()
    }));
  } else {
    die(Google::Ads::Common::AuthError->new({
      message => $response->message(),
      code => $response->code(),
      content => $response->content()
    }));
  }
}

sub __content_to_hash {
  return map {$_ =~ /^(.+?)=(.+)$/; $1 => $2} split(/\n/, shift);
}

return 1;

=pod

=head1 NAME

Google::Ads::Common::AuthToken

=head1 SYNOPSIS

 use Google::Ads::Common::AuthToken;

 my $token;
 eval {
   $token = Google::Ads::Common::AuthToken::get_token({
     email => "user@domain.com",
     password => "password",
     service => "adwords"
   });
 };
 if ($@) {
   # $@ will contain a string explaining why the token request failed.
 } else {
   # Make use of $token.
 }

=head1 DESCRIPTION

Google::Ads::Common::AuthToken exposes one static method, get_token().

=head1 METHODS

=head2 get_token

A static method that attempts to log in to Google's ClientLogin service with the
provided credentials. The credentials should correspond to an active Google
Account; more info on Google Accounts can be found at
L<https://www.google.com/accounts/ManageAccount>.

=head3 Parameters

A hash reference with two keys:
{
  # The service name to access.
  service => "adwords",
  # The email address of a Google Account.
  email => "user@domain.com",
  # The password for the Google Account.
  password => "password",
}

=head3 Returns

A string containing the auth token suitable for use with the API.

=head3 Exceptions

If the attempt to log in fails, either due to bad or missing credentials or an
HTTP error contacting the service, get_token() will die() with a descriptive
error message.

=head1 LICENSE AND COPYRIGHT

Copyright 2011 Google Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

=head1 AUTHOR

Jeffrey Posnick E<lt>api.jeffy at gmail.comE<gt>

David Torres E<lt>api.davidtorres at gmail.comE<gt>

=head1 REPOSITORY INFORMATION

 $Rev: $
 $LastChangedBy: $
 $Id: $

=cut
