# Copyright 2011, Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

package Google::Ads::Common::HTTPTransport;

use strict;
use version;
use base qw(SOAP::WSDL::Transport::HTTP);

# The following needs to be on one line because CPAN uses a particularly hacky
# eval() to determine module versions.
use Google::Ads::Common::Constants; our $VERSION = ${Google::Ads::Common::Constants::VERSION};

sub oauth_handler {
  my ($self, $handler) = @_;
  $self->{_oauth_handler} = $handler if (defined $handler);
  return $self->{_oauth_handler};
}

sub send_receive {
  my ($self, %parameters) = @_;
  my $handler = $self->oauth_handler;

  if ($handler->is_oauth_enabled()) {
    # At this point we prepare the request for OAuth.
    $parameters{endpoint} =
        $handler->get_protected_resource_url($parameters{endpoint});
  }

  return $self->SUPER::send_receive(%parameters);
}

return 1;

=pod

=head1 NAME

Google::Ads::Common::HTTPTransport - Specialization of
L<SOAP::WSDL::Transport::HTTP> transport class with added logic to handle OAuth.

=head1 DESCRIPTION

Provides a thin transport class on top of L<SOAP::WSDL::Transport::HTTP>  to add
support to OAuth.

=head1 ATTRIBUTES

=head2 auth_handler

Holds an instance of L<Google::Ads::Common::OAuthHandler>, which will use to
execute all the required OAuth logic.

=head1 METHODS

=head2 send_receive

Overrides L<SOAP::WSDL::Transport::HTTP> send_receive method to change the
endpoint in case OAuth is enabled, uses the L<Google::Ads::AdWords::Client>
configured OAuth handler to leverage all the OAuth signing logic.

=head3 Parameters

The same as the L<SOAP::WSDL::Transport::HTTP> send_receive method.

=head1 LICENSE AND COPYRIGHT

Copyright 2011 Google Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

=head1 AUTHOR

David Torres E<lt>api.davidtorres at gmail.comE<gt>

=head1 REPOSITORY INFORMATION

 $Rev: $
 $LastChangedBy: $
 $Id: $

=cut
