# Copyright 2011, Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

package Google::Ads::Common::OAuthHandlerInterface;

# Method to retrieve an authorization URL for the user to put in a
# browser and authorize a request token.
# Meant to be implemented by a concrete class, which should issue a request
# token and return a valid URL for the user to authorize the token.
# The implementor must save the request token in the token attribute for later
# use by the upgrade_token method.
# A callback URL can be passed to re-direct the user after the token is
# authorized.
sub get_authorization_url {
  my ($self, $callback) = @_;
  die "Needs to be implemented by subclass";
}

# Method to upgrade an authorized request token to an access token for later
# use when issuing requests against protected resources.
# The implementer of this method must re-use the request token, stored in
# the token attribute and send a request to upgrade it.
sub upgrade_token {
  my ($self, $verifier) = @_;
  die "Needs to be implemented by subclass";
}

# Method to generate a signed URL to access a given resource URL.
# The implementer of this method must use the consumer key/secret and
# token/token secret to properly sign and generate the URL.
sub get_protected_resource_url {
  my ($self, $url) = @_;
  die "Needs to be implemented by subclass";
}

sub is_oauth_enabled {
  my ($self) = @_;
  die "Needs to be implemented by subclass";
}

1;

=pod

=head1 NAME

Google::Ads::Common::OAuthHandlerInterface

=head1 DESCRIPTION

Abstract interface to be implemented by concrete OAuth handlers to define the
necessary subroutines to fully implement OAuth access to Ad APIs.

=head1 METHODS

=head2 get_authorization_url

Meant to be implemented by a concrete class, which should issue a request
token and return a valid URL for the user to authorize the token.
The implementor must save the request token in the token attribute for later
use by the upgrade_token method.
A callback URL can be passed to re-direct the user after the token is
authorized.

=head3 Parameters

A required callback URL to which the user will be redirected after authorizing
access. A special "oob" out-of-band value is allow to just print the verifier
in screen.

=head3 Returns

The URL for the user to authorize access to his account. The user first must
login in the account that want to grant access to.

=head2 upgrade_token

Method to upgrade an authorized request token to an access token for later
use by the get_protected_resource_url method.
The implementer of this method must re-use the request token generated by
the get_authorization_url method and store it for later use.

=head3 Parameters

The verifier returned to your callback page or printed out if using 'oob'
special callback URL.

=head2 get_protected_resource_url

Method to generate a signed URL meant to access a protected resource URL as
an Ads API endpoint.
The implementer of this method must use the consumer key/secret pair and
token/secret pair to properly sign and generate the URL.

=head3 Parameters

The protected URL to access.

=head3 Returns

A signed URL that can be used to make the request.


=head2 is_oauth_enabled

Should return true if the class is prepared to sign and return URLs to access
protected resources, which is most of the cases means the class holds a valid
consumer key/secret and token/secret pairs.

=head3 Returns

True if the class is prepared to sign URL to access protected resources, false
otherwise.

=head1 LICENSE AND COPYRIGHT

Copyright 2011 Google Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

     http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

=head1 AUTHOR

David Torres E<lt>api.davidtorres at gmail.comE<gt>

=head1 REPOSITORY INFORMATION

 $Rev: $
 $LastChangedBy: $
 $Id: $

=cut
