#!/usr/bin/perl
#
# Copyright 2011, Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Unit tests for the Google::Ads::AdWords::AuthToken module.
#
# Author: David Torres <api.davidtorres@gmail.com>

use strict;

use File::Basename;
use File::Spec;
use HTTP::Response;
use Test::More (tests => 11);
use Test::MockObject;

# Set up @INC at runtime with an absolute path.
my $lib_path = File::Spec->catdir(dirname($0), "..", "lib");
push(@INC, $lib_path);

use_ok("Google::Ads::Common::AuthToken");

# Test that missing parameters throw an appropriate error.
eval {
  Google::Ads::Common::AuthToken::get_token({});
};
like($@, qr/Required hash parameter 'service' not provided./,
     "Missing email param");

eval {
  Google::Ads::Common::AuthToken::get_token({
    email => "user\@domain.com",
  });
};
like($@, qr/Required hash parameter 'service' not provided./,
     "Missing password param");

my $user_agent_mock = Test::MockObject->new();
$user_agent_mock->set_true("env_proxy");
$user_agent_mock->set_true("agent");
$user_agent_mock->mock(request => sub {
  my $response = HTTP::Response->new(403, "Forbidden (code 403)");
  $response->content("Error=BadAuthentication");
  return $response;
});
Google::Ads::Common::AuthToken::set_user_agent($user_agent_mock);

eval {
  Google::Ads::Common::AuthToken::get_token({
    email => "user\@domain.com",
    password => "dummypassword",
    service => "adwords"
  });
};
isa_ok($@, "Google::Ads::Common::AuthError", "Authentication error");
is($@->get_code(), 403, "Error 403 code");
is($@->get_message(), "Forbidden (code 403)", "Error message");
is($@->get_content(), "Error=BadAuthentication", "Error content");

$user_agent_mock->mock(request => sub {
  my $response = HTTP::Response->new(403, "Forbidden (code 403)");
  $response->content("Error=CaptchaRequired\nCaptchaToken=123\nCaptchaUrl=" .
                     "captcha_image_url\nUrl=url");
  return $response;
});

eval {
  Google::Ads::Common::AuthToken::get_token({
    email => "user\@domain.com",
    password => "dummypassword",
    service => "adwords",
    server => "myserver"
  });
};
isa_ok($@, "Google::Ads::Common::CaptchaRequiredError", "Captcha error");
is($@->get_token(), "123", "Captcha token");
is($@->get_image(), "myserver/accounts/captcha_image_url", "Captcha image");
is($@->get_url(), "url", "Captcha URL");
