#!/usr/bin/perl
#
# Copyright 2011, Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Unit (not functional) tests for the Google::Ads::AdWords::Client module.
# Functional tests of the various AdWords API services will be performed in a
# separate test.
#
# Author: David Torres <api.davidtorres@gmail.com>

use strict;

use File::Basename;
use File::Spec;
use Test::More (tests => 12);

# Set up @INC at runtime with an absolute path.
my $lib_path = File::Spec->catdir(dirname($0), "..", "lib");
push(@INC, $lib_path);

# Test Google::Ads::AdWords::Client module should be included before any
# other Google::Ads module.
{
  $INC{"Google/Ads/Fake/Module/Load.pm"} = 'for testing only';
  my $warnings_found;
  local $SIG{__WARN__} = sub {
    my $warning = shift;
    my $expected = "Google::Ads::AdWords::Client should be loaded before " .
        "other Google::Ads::";
    like $warning, qr/$expected/,
        "Warnings for modules loaded in the wrong order"
        and $warnings_found = 1;
  };

  # Testing is ok to use the Client class
  use_ok("Google::Ads::AdWords::Client")
      or die "Cannot load 'Google::Ads::AdWords::Client'";

  if (not $warnings_found) {
    fail "A warning for module load order should have occurred, but did not";
  }
}

# Test client initialization, including reading from properties files.
my $properties_file =
    File::Spec->catdir(dirname($0), qw(testdata client.test.input));
my $password = "password_overide";
my $client = Google::Ads::AdWords::Client->new({
  password => $password,
  properties_file => $properties_file,
});
is($client->get_email(), "user\@domain.com", "Basic properties file reading");
is($client->get_password(), $password, "Local param overrides properties file");
is($client->get_client_id(), "client_1+user\@domain.com", "Read of client id");
is($client->get_user_agent(),
   "AwApi-Perl-${Google::Ads::AdWords::Constants::VERSION}|perl-unit-tests",
   "Read of user agent and appending of version");
is($client->get_developer_token(), "dev-token",
   "Read of developer token");
is($client->get_auth_token(), "auth-token",
   "Read of auth token");
is($client->get_alternate_url(), "https://adwords-sandbox.google.com",
   "Read of alternate url");

# Test basic get/set methods.
$client->set_die_on_faults(1);
is($client->get_die_on_faults(), 1, "get/set die_on_faults()");

# Make sure this supports all the services we think it should for each version.
$client->set_version("v201109");
my @services = qw(AdExtensionOverrideService AdGroupAdService
                  AdGroupCriterionService AdGroupService AdParamService
                  AlertService BulkMutateJobService
                  CampaignAdExtensionService CampaignCriterionService
                  CampaignService CampaignTargetService ConversionTrackerService
                  ConstantDataService CustomerSyncService DataService
                  ExperimentService GeoLocationService InfoService
                  LocationCriterionService MediaService MutateJobService
                  ReportDefinitionService ServicedAccountService
                  TargetingIdeaService TrafficEstimatorService
                  UserListService);
can_ok($client, @services);

ok(Google::Ads::AdWords::Client->new && Google::Ads::AdWords::Client->new,
   "Can construct more than one client object.");
